/*
 * Created on 11 jun 2009
 */

package jawnae.js;

import java.util.Map;
import java.util.Map.Entry;

import jawnae.js.x.JsDebugRegistry;

import org.mozilla.javascript.Context;
import org.mozilla.javascript.ContextFactory;
import org.mozilla.javascript.Function;
import org.mozilla.javascript.ImporterTopLevel;
import org.mozilla.javascript.Scriptable;
import org.mozilla.javascript.ScriptableObject;

public class JsScope
{
   public final Context    rhinoContext;
   public final Scriptable rhinoScope;

   public JsScope()
   {
      this((JsScope) null);
   }

   public JsScope(JsScope scope)
   {
      if (scope == null)
      {
         // ensure unshared context
         if (Context.getCurrentContext() != null)
         {
            throw new IllegalStateException("another context is already active");
         }

         this.rhinoContext = Context.enter();

         Scriptable prototype = this.rhinoContext.initStandardObjects();
         Scriptable topLevel = new ImporterTopLevel(this.rhinoContext);
         prototype.setParentScope(topLevel);

         this.rhinoScope = this.rhinoContext.newObject(prototype);
         this.rhinoScope.setPrototype(prototype);

         this.exitContext();
      }
      else
      {
         scope.enterContext();

         Scriptable prototype = scope.rhinoScope.getPrototype();
         this.rhinoContext = scope.rhinoContext;
         this.rhinoScope = this.rhinoContext.newObject(prototype);
         this.rhinoScope.setPrototype(prototype);
         this.rhinoScope.getPrototype();

         scope.exitContext();
      }

      if (scope != null)
      {
         this.rhinoScope.setParentScope(scope.rhinoScope);
         //  this.rhinoScope.setPrototype(null);
      }

      if (this.rhinoContext.getOptimizationLevel() == 0)
      {
         this.rhinoContext.setOptimizationLevel(9);
      }
   }

   // required for JsDebugger

   JsScope(Context context, Scriptable scope)
   {
      if (context == null)
         throw new NullPointerException();
      if (scope == null)
         throw new NullPointerException();
      this.rhinoContext = context;
      this.rhinoScope = scope;
   }

   // even more convenience methods

   public JsScript addScript(String code)
   {
      return new JsScript(code, this);
   }

   public JsScript addScript(String code, String sourceLocation)
   {
      return new JsScript(code, this, sourceLocation, 1);
   }

   public JsScript addScript(String code, String sourceLocation, int lineOffset)
   {
      return new JsScript(code, this, sourceLocation, lineOffset);
   }

   public JsFunction addFunction(String code)
   {
      return new JsFunction(code, this);
   }

   public JsFunction addFunction(String code, String sourceLocation)
   {
      return new JsFunction(code, this, sourceLocation, 1);
   }

   public JsFunction addFunction(String code, String sourceLocation, int lineOffset)
   {
      return new JsFunction(code, this, sourceLocation, lineOffset);
   }

   //

   public Context enterContext()
   {
      Context context = ContextFactory.getGlobal().enterContext(this.rhinoContext);

      if (context != this.rhinoContext)
      {
         throw new IllegalStateException("scope context[" + this.rhinoContext + "] != entered content[" + context + "]");
      }

      if (context.getOptimizationLevel() == -1)
      {
         JsDebugRegistry.onContextOpened(this.rhinoContext, this);
      }

      return context;
   }

   public void exitContext()
   {
      Context context = Context.getCurrentContext();

      if (context == null)
      {
         throw new IllegalStateException("current context == null");
      }

      if (context != this.rhinoContext)
      {
         throw new IllegalStateException("scope context[" + this.rhinoContext + "] != current content[" + context + "]");
      }

      Context.exit();

      if (Context.getCurrentContext() == null)
      {
         if (context.getOptimizationLevel() == -1)
         {
            JsDebugRegistry.onContextClosed(this.rhinoContext);
         }
      }
   }

   //

   public Object eval(String code)
   {
      try
      {
         Context context = this.enterContext();

         return context.evaluateString(this.rhinoScope, code, "<runtime>", 0, null);
      }
      finally
      {
         this.exitContext();
      }
   }

   //

   public void put(String name, Object obj)
   {
      try
      {
         this.enterContext();

         ScriptableObject.putProperty(this.rhinoScope, name, Context.javaToJS(obj, rhinoScope));
      }
      finally
      {
         this.exitContext();
      }
   }

   public void putAll(Map<String, Object> map)
   {
      try
      {
         this.enterContext();

         for (Entry<String, Object> entry : map.entrySet())
         {
            this.put(entry.getKey(), entry.getValue());
         }
      }
      finally
      {
         this.exitContext();
      }
   }

   public boolean has(String name)
   {
      try
      {
         this.enterContext();

         return ScriptableObject.hasProperty(this.rhinoScope, name);
      }
      finally
      {
         this.exitContext();
      }
   }

   public String[] getProperties()
   {
      try
      {
         this.enterContext();

         Object[] ids = ScriptableObject.getPropertyIds(this.rhinoScope);
         String[] names = new String[ids.length];
         for (int i = 0; i < names.length; i++)
            names[i] = (String) ids[i];
         return names;
      }
      finally
      {
         this.exitContext();
      }
   }

   public void printProperties()
   {
      this.printProperties("this");
   }

   public void printProperties(String name)
   {
      String[] props = this.getProperties();
      System.out.println(name + ".propCount=" + props.length);
      for (String prop : props)
      {
         Object val = this.getJs(prop);
         if (val instanceof Function)
            val = "[[function]]";
         else if (val instanceof Scriptable)
            val = "[[script | scope]]";
         System.out.println(name + "." + prop + "=" + val);
      }
   }

   public <T> T getVariable(String name, Class<T> type)
   {
      try
      {
         this.enterContext();

         return (T) Context.jsToJava(this.getJs(name), type);
      }
      finally
      {
         this.exitContext();
      }
   }

   public Object getJs(String name)
   {
      try
      {
         this.enterContext();

         return ScriptableObject.getProperty(this.rhinoScope, name);
      }
      finally
      {
         this.exitContext();
      }
   }

   public JsScope getAsScope(String name)
   {
      return new JsScope(this.rhinoContext, (Scriptable) this.getJs(name));
   }

   // convenient

   public boolean getBoolean(String name)
   {
      return ((Boolean) this.getVariable(name, Boolean.class)).booleanValue();
   }

   public int getInt(String name)
   {
      return ((Integer) this.getVariable(name, Integer.class)).intValue();
   }

   public long getLong(String name)
   {
      return ((Long) this.getVariable(name, Long.class)).longValue();
   }

   public float getFloat(String name)
   {
      return ((Float) this.getVariable(name, Float.class)).floatValue();
   }

   public double getDouble(String name)
   {
      return ((Double) this.getVariable(name, Double.class)).doubleValue();
   }

   public String getString(String name)
   {
      return (String) this.getVariable(name, String.class);
   }

   //

   public void putBoolean(String name, boolean val)
   {
      this.put(name, Boolean.valueOf(val));
   }

   public void putInt(String name, int val)
   {
      this.put(name, Integer.valueOf(val));
   }

   public void putLong(String name, long val)
   {
      this.put(name, Long.valueOf(val));
   }

   public void putFloat(String name, float val)
   {
      this.put(name, Float.valueOf(val));
   }

   public void putDouble(String name, double val)
   {
      this.put(name, Double.valueOf(val));
   }

   public void putString(String name, String val)
   {
      this.put(name, val);
   }

   public void putFunction(String name, JsFunction function)
   {
      this.put(name, function.rhinoFunction);
   }

   public void putFunction(JsFunction function)
   {
      this.putFunction(function.getName(), function);
   }

   //

   public void setInterpretedContext(boolean interpreted)
   {
      this.setOptLevel(interpreted ? -1 : 9);
   }

   private void setOptLevel(int level)
   {
      Context context = this.enterContext();

      try
      {
         context.setOptimizationLevel(level);
      }
      finally
      {
         this.exitContext();
      }
   }

   //

   private String name;

   public void setName(String name)
   {
      this.name = name;
   }

   public String getName()
   {
      return this.name;
   }

   //

   public Context fetchRhinoContext()
   {
      return this.rhinoContext;
   }

   public Scriptable fetchRhinoScope()
   {
      return this.rhinoScope;
   }

   //

   @Override
   public String toString()
   {
      return "JsScope[" + this.name + "]";
   }
}